part of tags;

/// A function to bind to an element's [model].
typedef void BindModel(HasElement model);

/// A function to build children of a [Tag].
///
/// Each tag function that is called from within a build function will append
/// its element as a child of the current [buildHostNode].  The order of such 
/// calls is equal to the order in which the children are appended.
typedef void BuildChildren();

/// A function to create an instance of a node.
///
/// All tag functions that are called from within a create function will _not_
/// append any elements to the node being constructed.  The [buildHostNode] will 
/// be _null_ within the scope of this function.  This allows the side-effect 
/// free usage of [Tag]s in class initializers.
typedef Node CreateNode(args);

/// A function to build an html element.
///
/// An optional build function may be provided to build child elements.
typedef E Tag<E extends Element>([BuildChildren build, args]);

// Stack of nodes used to build trees.  
// The top of the stack is always the host for the current build scope.
final _scopes = new List<Node>();

// Whether or not the builder is currently in a [Build] function scope.
bool _isBuilding = false;

/// The node currently being built; will be `null` if none or within the scope
/// of a [CreateNode] function.
Node get buildHostNode {
  if (!_isBuilding || _scopes.isEmpty) {
    return null;
  } else {
    return _scopes.last;
  }
}

/// Append the given [node] to the current [buildHostNode].
///
/// An optional build function may be provided to build child elements.  This 
/// function provides a means to append existing elements in a tag builder 
/// expression.
Node append(Node node, [BuildChildren b]) => _buildNode((_) => node, b);

/// Append the given [text] as a text node to the current [buildHostNode].
///
/// This function may only be called from inside a [BuildChildren] function and 
/// the [buildHostNode] must be an element.
void appendText(String text) {
  final _host = buildHostNode;
  if (_host is! Element) 
    throw new StateError('text() may only be called if host is an Element.');
  _host.appendText(text);
}

/// Call [binder] with the given [model] and return the model's element.
Element bindModel(HasElement model, BindModel binder) {
  if (binder != null) binder(model);
  return model.element;
}

/// Create a node and build its children.
///
/// The node constructed by [createNode] will be the [buildHostNode] when the 
/// [buildChildren] function is called.  All [Tag] functions that are called 
/// within the scope of [buildChildren] will append their element as a child of 
/// the [buildHostNode].  [Tag] functions called within the scope of 
/// [createNode] are side-effect free.
Node _buildNode(CreateNode createNode, [BuildChildren buildChildren, args]) {
  final bool outerIsBuilding = _isBuilding;
  var node;
  try {
    _isBuilding = false;
    node = createNode(args);
    assert(node != null);
  } finally {  
    _isBuilding = outerIsBuilding;  
  }
  if (outerIsBuilding && !_scopes.isEmpty) {
    _scopes.last.append(node);  
  }
  if (buildChildren == null) {
    return node;
  }
  _scopes.add(node);  
  try {
    _isBuilding = true;
    buildChildren();
  } finally {
    _isBuilding = outerIsBuilding;
    _scopes.removeLast();
  }
  return node;
}

/// Creates a new [Tag] with the given [createNode] function and [args].
Tag createTag(CreateNode createNode, [args]) => 
    ([BuildChildren b, args]) => _buildNode(createNode, b, args);

/// Create a shadow root for the current [buildHostNode].
///
/// This function may only be called from inside a [BuildChildren] function and 
/// the host must be an element.
ShadowRoot shadow([BuildChildren b]) {
  final _host = buildHostNode;
  if (_host is! Element) {
    throw new StateError('shadow() may only be called if host is an Element.');
  }
  return _buildNode((_) => _host.createShadowRoot(), b);
}

/// An object that has an element.
abstract class HasElement {
  Element get element;
}

